#ifndef GCSA_H
#define GCSA_H

#include <fstream>
#include <vector>

#include <bits/deltavector.h>
#include <bits/rlevector.h>

#include "graph.h"


namespace CSA
{


class GCSA
{
  public:

//--------------------------------------------------------------------------
//  CONSTRUCTION
//--------------------------------------------------------------------------

    const static usint ARRAY_BLOCK_SIZE = 32;
    const static usint CHILDREN_BLOCK_SIZE = 32;
    const static usint OUTGOING_BLOCK_SIZE = 32;
    const static usint SAMPLE_BLOCK_SIZE = 32;
    const static usint SAMPLE_RATE = 16;

    const static usint ALPHABET_SIZE = 5;
    const static std::string ALPHABET;
    const static std::string COMPLEMENT;

    GCSA(const std::string& base_name);
    GCSA(PathGraph& graph, bool print = false);
    ~GCSA();

    void writeTo(const std::string& base_name) const;

    bool isOk() const;

//--------------------------------------------------------------------------
//  REPORTING
//--------------------------------------------------------------------------

    // Returns the size of the data structure.
    usint reportSize(bool print = false) const;

    usint getSize() const;

//--------------------------------------------------------------------------
//  SEARCHING
//--------------------------------------------------------------------------

    // Returns the SA range for suffixes starting with the pattern.
    pair_type find(const std::string& pattern, bool reverse_complement = false) const;

    // Returns the set of unique node values for the given BWT range of nodes.
    // The user must delete the vector.
    // Returns 0 in case of errors.
    std::vector<usint>* locate(pair_type range) const;

//--------------------------------------------------------------------------
//  INTERNAL VARIABLES
//--------------------------------------------------------------------------

  private:
    bool ok;

    DeltaVector* array[ALPHABET_SIZE];    // BWT
    RLEVector* first_children;            // F
    RLEVector* outgoing;                  // M

    DeltaVector* sampled_positions;
    ReadBuffer* samples;

    usint counts[ALPHABET_SIZE];
    usint cumulative[ALPHABET_SIZE + 1];  // C

//--------------------------------------------------------------------------
//  INTERNAL STUFF
//--------------------------------------------------------------------------

    pair_type getIndexRange(usint c, bool complement = false) const;
    usint getCharacter(usint index) const;  // Parameter should be a rank of an outgoing edge.
    pair_type LF(pair_type range, usint c, bool complement = false) const;
    pair_type convertToNodeRange(pair_type edge_range) const;
    pair_type Psi(pair_type range) const; // Follows the first outgoing edge.

    void fillArrayC();

    // These are not allowed.
    GCSA();
    GCSA(const GCSA&);
    GCSA& operator = (const GCSA&);
};


} // namespace CSA


#endif // GCSA_H
